import dataclasses
from inspect import cleandoc
from typing import Any, Dict, Generic, Optional, Type, TypeVar

import django
import django.db.models
import strawberry
from strawberry import UNSET
from strawberry.annotation import StrawberryAnnotation
from strawberry.field import UNRESOLVED

from . import utils
from .fields.field import StrawberryDjangoField
from .fields.types import (
    get_model_field,
    is_optional,
    resolve_model_field_name,
    resolve_model_field_type,
)
from .settings import strawberry_django_settings as django_settings


_type = type


def get_type_attr(type_, field_name: str):
    attr = getattr(type_, field_name, UNSET)
    if attr is UNSET:
        attr = getattr(type_, "__dataclass_fields__", {}).get(field_name, UNSET)
    return attr


def get_field(
    django_type: "StrawberryDjangoType",
    field_name: str,
    field_annotation: Optional[StrawberryAnnotation] = None,
):
    attr = get_type_attr(django_type.origin, field_name)

    if utils.is_field(attr):
        field = django_type.field_cls.from_field(attr, django_type)
    else:
        field = django_type.field_cls(
            default=attr,
            type_annotation=field_annotation,
        )

    field.python_name = field_name
    if field_name in django_type.origin.__dict__.get("__annotations__", {}):
        # store origin django type for further usage
        field.origin_django_type = django_type

    if field_annotation:
        # annotation of field is used as a class type
        field.type_annotation = field_annotation
        field.is_auto = utils.is_auto(field_annotation)

    try:
        # resolve the django_name and check if it is relation field. django_name
        # is used to access the field data in resolvers
        django_name = field.django_name or field_name
        model_field = get_model_field(django_type.model, django_name)
        field.django_name = resolve_model_field_name(
            model_field, django_type.is_input, django_type.is_filter
        )
        field.is_relation = model_field.is_relation

        # Use the Django field help_text if no other description is available.
        settings = django_settings()
        if not field.description and settings["FIELD_DESCRIPTION_FROM_HELP_TEXT"]:
            model_field_help_text = getattr(model_field, "help_text", "")
            field.description = str(model_field_help_text) or None
    except django.core.exceptions.FieldDoesNotExist:
        if field.django_name or field.is_auto:
            raise  # field should exist, reraise caught exception
        model_field = None

    if field.is_relation:
        # change relation field type to auto if field is inherited from another
        # type. for example if field is inherited from output type but we are
        # configuring field for input type
        if not utils.is_similar_django_type(django_type, field.origin_django_type):
            field.is_auto = True

    # Only set the type_annotation for auto fields if they don't have a base_resolver.
    # Since strawberry 0.139 the type_annotation has a higher priority than the
    # resolver's annotation, and that would force our automatic model resolution to be
    # used instead of the resolver's type annotation.
    if field.is_auto and not field.base_resolver:
        # resolve type of auto field
        field_type = resolve_model_field_type(model_field, django_type)
        field.type_annotation = StrawberryAnnotation(field_type)

    if field.type_annotation and is_optional(
        model_field, django_type.is_input, django_type.is_partial
    ):
        field.type_annotation.annotation = Optional[field.type_annotation.annotation]

    if django_type.is_input:
        if field.default is dataclasses.MISSING:
            # strawberry converts UNSET value to MISSING, let's set
            # it back to UNSET. this is important especially for partial
            # input types
            # TODO: could strawberry support UNSET default value?
            field.default_value = UNSET
            field.default = UNSET

    return field


def get_fields(django_type: "StrawberryDjangoType"):
    annotations = utils.get_annotations(django_type.origin)
    fields: Dict[str, StrawberryDjangoField] = {}

    # collect all annotated fields
    for field_name, field_annotation in annotations.items():
        field = get_field(django_type, field_name, field_annotation)
        fields[field_name] = field

    # collect non-annotated strawberry fields
    for field_name in dir(django_type.origin):
        if field_name in fields:
            continue
        attr = getattr(django_type.origin, field_name)
        if not utils.is_strawberry_field(attr):
            continue
        field = get_field(django_type, field_name)
        fields[field_name] = field

    return list(fields.values())


_O = TypeVar("_O", bound=type)
_M = TypeVar("_M", bound=django.db.models.Model)


@dataclasses.dataclass
class StrawberryDjangoType(Generic[_O, _M]):
    origin: _O
    model: Type[_M]
    is_input: bool
    is_partial: bool
    is_filter: bool
    filters: Any
    order: Any
    pagination: Any
    field_cls: Type[StrawberryDjangoField]


def process_type(
    cls,
    model: Type[django.db.models.Model],
    *,
    filters=UNSET,
    pagination=UNSET,
    order=UNSET,
    field_cls=UNSET,
    **kwargs
):
    original_annotations = cls.__dict__.get("__annotations__", {})

    if not field_cls or field_cls is UNSET:
        field_cls = StrawberryDjangoField

    django_type = StrawberryDjangoType(
        origin=cls,
        model=model,
        is_input=kwargs.get("is_input", False),
        is_partial=kwargs.pop("partial", False),
        is_filter=kwargs.pop("is_filter", False),
        filters=filters,
        order=order,
        pagination=pagination,
        field_cls=field_cls,
    )

    fields = get_fields(django_type)

    # update annotations and fields
    cls.__annotations__ = cls_annotations = {}
    for field in fields:
        annotation = None

        if field.type_annotation and field.type_annotation.annotation:
            annotation = field.type_annotation.annotation
        elif field.base_resolver and field.base_resolver.type_annotation:
            annotation = field.base_resolver.type_annotation.annotation

        # UNRESOLVED is not a valid annotation, it is just an indication that the type
        # could not be resolved. In this case just fallback to None
        if annotation is UNRESOLVED:
            annotation = None

        # TODO: should we raise an error if annotation is None here?

        cls_annotations[field.name] = annotation
        setattr(cls, field.name, field)

    # Strawberry >= 0.92.0 defines `is_type_of` for types implementing
    # interfaces if the attribute has not been set yet. It allows only
    # instances of `cls` to be returned, we should allow model instances
    # too.
    if not hasattr(cls, "is_type_of"):
        cls.is_type_of = lambda obj, _info: isinstance(obj, (cls, model))

    # Get type description from either kwargs, or the model's docstring
    settings = django_settings()
    description = kwargs.pop("description", None)
    if not description and settings["TYPE_DESCRIPTION_FROM_MODEL_DOCSTRING"]:
        description = cleandoc(model.__doc__) or None

    strawberry.type(cls, description=description, **kwargs)

    # restore original annotations for further use
    cls.__annotations__ = original_annotations
    cls._django_type = django_type

    return cls


def type(model, *, filters=UNSET, **kwargs):
    def wrapper(cls):
        return process_type(cls, model, filters=filters, **kwargs)

    return wrapper


def input(model, *, partial=False, **kwargs):
    return type(model, partial=partial, is_input=True, **kwargs)


def mutation(model, **kwargs):
    return type(model, **kwargs)



from strawberry.schema.execute import validate_document

from .django_cache_base import DjangoCacheBase


class DjangoValidationCache(DjangoCacheBase):
    def on_validation_start(self) -> None:
        execution_context = self.execution_context

        errors = self.execute_cached(
            validate_document,
            execution_context.schema._schema,
            execution_context.graphql_document,
            execution_context.validation_rules,
        )
        execution_context.errors = errors



from . import utils
from .type import type as strawberry_django_type


def from_type(type_, *, is_input=False, partial=False):
    type_name = type_.__name__
    if partial:
        type_name += "Partial"
    if is_input:
        type_name += "Input"
    type_name += "Type"
    model = utils.get_django_model(type_)
    cls = type(type_name, (type_,), {})
    return strawberry_django_type(model, is_input=is_input, partial=partial)(cls)



import warnings
from typing import Any, Dict

from strawberry import auto as _deprecated_auto  # noqa: F401

from . import auth, filters, mutations, ordering, types
from .fields.field import field
from .fields.types import (  # noqa: F401
    DjangoFileType,
    DjangoImageType,
    DjangoModelType,
    ManyToManyInput,
    ManyToOneInput,
    OneToManyInput,
    OneToOneInput,
    is_auto as _deprecated_is_auto,
)
from .filters import filter_deprecated as filter
from .resolvers import django_resolver
from .type import input, mutation, type
from .utils import fields


_deprecated_names: Dict[str, str] = {
    "auto": (
        "importing `auto` from `strawberry_django` is deprecated, "
        "import instead from `strawberry` directly."
    ),
    "is_auto": (
        "`is_auto` is deprecated use `isinstance(value, StrawberryAuto)` instead."
    ),
}


def __getattr__(name: str) -> Any:
    if name in _deprecated_names:
        warnings.warn(_deprecated_names[name], DeprecationWarning, stacklevel=2)
        return globals()[f"_deprecated_{name}"]

    raise AttributeError(f"module {__name__} has no attribute {name}")


__all__ = [
    "auth",
    "filters",
    "filter",
    "ordering",
    "types",
    "field",
    "DjangoFileType",
    "DjangoImageType",
    "DjangoModelType",
    "OneToOneInput",
    "OneToManyInput",
    "ManyToOneInput",
    "ManyToManyInput",
    "fields",
    "mutations",
    "django_resolver",
    "type",
    "input",
    "mutation",
]



import datetime
import decimal
import uuid
from typing import TYPE_CHECKING, Any, List, Optional, Type, Union

import django
import strawberry
from django.db.models import Field, Model, fields
from django.db.models.fields.reverse_related import ForeignObjectRel
from strawberry import UNSET
from strawberry.auto import StrawberryAuto
from strawberry.scalars import JSON

from .. import filters


if TYPE_CHECKING:
    from strawberry_django.type import StrawberryDjangoType


@strawberry.type
class DjangoFileType:
    name: str
    path: str
    size: int
    url: str


@strawberry.type
class DjangoImageType(DjangoFileType):
    width: int
    height: int


@strawberry.type
class DjangoModelType:
    pk: strawberry.ID


@strawberry.input
class OneToOneInput:
    set: Optional[strawberry.ID]


@strawberry.input
class OneToManyInput:
    set: Optional[strawberry.ID]


@strawberry.input
class ManyToOneInput:
    add: Optional[List[strawberry.ID]] = UNSET
    remove: Optional[List[strawberry.ID]] = UNSET
    set: Optional[List[strawberry.ID]] = UNSET


@strawberry.input
class ManyToManyInput:
    add: Optional[List[strawberry.ID]] = UNSET
    remove: Optional[List[strawberry.ID]] = UNSET
    set: Optional[List[strawberry.ID]] = UNSET


field_type_map = {
    fields.AutoField: strawberry.ID,
    fields.BigAutoField: strawberry.ID,
    fields.BigIntegerField: int,
    fields.BooleanField: bool,
    fields.CharField: str,
    fields.DateField: datetime.date,
    fields.DateTimeField: datetime.datetime,
    fields.DecimalField: decimal.Decimal,
    fields.EmailField: str,
    fields.files.FileField: DjangoFileType,
    fields.FilePathField: str,
    fields.FloatField: float,
    fields.files.ImageField: DjangoImageType,
    fields.GenericIPAddressField: str,
    fields.IntegerField: int,
    fields.NullBooleanField: Optional[bool],
    fields.PositiveIntegerField: int,
    fields.PositiveSmallIntegerField: int,
    fields.SlugField: str,
    fields.SmallAutoField: strawberry.ID,
    fields.SmallIntegerField: int,
    fields.TextField: str,
    fields.TimeField: datetime.time,
    fields.URLField: str,
    fields.UUIDField: uuid.UUID,
    fields.related.ForeignKey: DjangoModelType,
    fields.reverse_related.ManyToOneRel: List[DjangoModelType],
    fields.related.OneToOneField: DjangoModelType,
    fields.reverse_related.OneToOneRel: DjangoModelType,
    fields.related.ManyToManyField: List[DjangoModelType],
    fields.reverse_related.ManyToManyRel: List[DjangoModelType],
}

if django.VERSION >= (3, 1):
    field_type_map.update(
        {
            fields.json.JSONField: JSON,
            fields.PositiveBigIntegerField: int,
        }
    )

input_field_type_map = {
    fields.files.FileField: NotImplemented,
    fields.files.ImageField: NotImplemented,
    fields.related.ForeignKey: OneToManyInput,
    fields.reverse_related.ManyToOneRel: ManyToOneInput,
    fields.related.OneToOneField: OneToOneInput,
    fields.reverse_related.OneToOneRel: OneToOneInput,
    fields.related.ManyToManyField: ManyToManyInput,
    fields.reverse_related.ManyToManyRel: ManyToManyInput,
}


def resolve_model_field_type(
    model_field: Union[Field, ForeignObjectRel], django_type: "StrawberryDjangoType"
):
    model_field_type = type(model_field)
    field_type: Any = None
    if django_type.is_filter and model_field.is_relation:
        field_type = filters.DjangoModelFilterInput
    elif django_type.is_input:
        field_type = input_field_type_map.get(model_field_type, None)
    if field_type is None:
        field_type = field_type_map.get(model_field_type, NotImplemented)
    if field_type is NotImplemented:
        raise NotImplementedError(
            f"GraphQL type for model field '{model_field}' has not been implemented"
        )
    if django_type.is_filter == "lookups":
        # TODO: could this be moved into filters.py
        if not model_field.is_relation and field_type is not bool:
            field_type = filters.FilterLookup[field_type]
    return field_type


def resolve_model_field_name(
    model_field: Union[Field, ForeignObjectRel], is_input=False, is_filter=False
):
    if isinstance(model_field, ForeignObjectRel):
        return model_field.get_accessor_name()
    if is_input and not is_filter:
        return model_field.attname
    else:
        return model_field.name


def get_model_field(model: Type[Model], field_name: str):
    try:
        return model._meta.get_field(field_name)
    except django.core.exceptions.FieldDoesNotExist as e:
        model_field_names = []

        # we need to iterate through all the fields because reverse relation
        # fields cannot be accessed by get_field method
        for field in model._meta.get_fields():
            model_field_name = resolve_model_field_name(field)
            if field_name == model_field_name:
                return field
            model_field_names.append(model_field_name)

        e.args = (
            "{}, did you mean {}?".format(
                e.args[0], ", ".join([f"'{n}'" for n in model_field_names])
            ),
        )
        raise e


def is_auto(type_):
    return isinstance(type_, StrawberryAuto)


def is_optional(model_field, is_input, partial):
    if partial:
        return True
    if not model_field:
        return False
    if is_input:
        if isinstance(model_field, fields.AutoField):
            return True
        if isinstance(model_field, fields.reverse_related.OneToOneRel):
            return model_field.null
        if model_field.many_to_many or model_field.one_to_many:
            return True
        has_default = model_field.default is not fields.NOT_PROVIDED
        if model_field.blank or has_default:
            return True
    if not isinstance(
        model_field,
        (fields.reverse_related.ManyToManyRel, fields.reverse_related.ManyToOneRel),
    ) or isinstance(model_field, fields.reverse_related.OneToOneRel):
        # OneToOneRel is the subclass of ManyToOneRel, so additional check is needed
        return model_field.null
    return False



"""
Code for interacting with Django settings.
"""
try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict

from django.conf import settings


class StrawberryDjangoSettings(TypedDict):
    """
    Dictionary defining the shape `settings.STRAWBERRY_DJANGO` should have.

    All settings are optional and have defaults as described in their docstrings and
    defined in `DEFAULT_DJANGO_SETTINGS`.
    """

    FIELD_DESCRIPTION_FROM_HELP_TEXT: bool
    """(Default: False) If True, field descriptions will be fetched from the
    corresponding model field's `help_text` attribute."""

    TYPE_DESCRIPTION_FROM_MODEL_DOCSTRING: bool
    """(Default: False) If True, type descriptions will be fetched from the
    corresponding model's docstring."""


DEFAULT_DJANGO_SETTINGS = StrawberryDjangoSettings(
    FIELD_DESCRIPTION_FROM_HELP_TEXT=False,
    TYPE_DESCRIPTION_FROM_MODEL_DOCSTRING=False,
)


def strawberry_django_settings() -> StrawberryDjangoSettings:
    """
    Return the dictionary from `settings.STRAWBERRY_DJANGO`, with defaults for missing
    keys.

    Preferred to direct access for the type hints and defaults.
    """
    defaults = DEFAULT_DJANGO_SETTINGS
    customized = {**defaults, **getattr(settings, "STRAWBERRY_DJANGO", {})}
    return customized



import enum
from typing import List, Optional, Type

import strawberry
from django.db.models import QuerySet
from strawberry import UNSET
from strawberry.arguments import StrawberryArgument
from strawberry.auto import StrawberryAuto
from strawberry.types import Info

from strawberry_django.utils import fields

from . import utils
from .arguments import argument


@strawberry.enum
class Ordering(enum.Enum):
    ASC = "ASC"
    DESC = "DESC"


def generate_order_args(order, prefix=""):
    args = []
    for field in fields(order):
        ordering = getattr(order, field.name, UNSET)
        if ordering is UNSET:
            continue
        if ordering == Ordering.ASC:
            args.append(f"{prefix}{field.name}")
        elif ordering == Ordering.DESC:
            args.append(f"-{prefix}{field.name}")
        else:
            subargs = generate_order_args(ordering, prefix=f"{prefix}{field.name}__")
            args.extend(subargs)
    return args


def order(model):
    def wrapper(cls):
        for name, type_ in cls.__annotations__.items():
            if isinstance(type_, StrawberryAuto):
                type_ = Ordering
            cls.__annotations__[name] = Optional[type_]
            setattr(cls, name, UNSET)
        return strawberry.input(cls)

    return wrapper


def apply(order, queryset: QuerySet) -> QuerySet:
    if order is UNSET or order is None:
        return queryset
    args = generate_order_args(order)
    if not args:
        return queryset
    return queryset.order_by(*args)


class StrawberryDjangoFieldOrdering:
    def __init__(self, order=UNSET, **kwargs):
        self.order = order
        super().__init__(**kwargs)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        arguments = []
        if not self.base_resolver:
            order = self.get_order()
            if order and order is not UNSET and self.is_list:
                arguments.append(argument("order", order))
        return super().arguments + arguments

    def get_order(self) -> Optional[Type]:
        if self.order is not UNSET:
            return self.order
        type_ = utils.unwrap_type(self.type or self.child.type)

        if utils.is_django_type(type_):
            return type_._django_type.order
        return None

    def apply_order(self, queryset: QuerySet, order) -> QuerySet:
        return apply(order, queryset)

    def get_queryset(
        self, queryset: QuerySet, info: Info, order: Type = UNSET, **kwargs
    ):
        queryset = super().get_queryset(queryset, info, **kwargs)
        return self.apply_order(queryset, order)



from typing import Any, Optional, Type, TypeVar

from django.db import models
from strawberry import UNSET
from strawberry.annotation import StrawberryAnnotation
from strawberry.field import StrawberryField
from strawberry.type import StrawberryList, StrawberryOptional
from strawberry.types import Info

from .. import utils
from ..filters import StrawberryDjangoFieldFilters
from ..ordering import StrawberryDjangoFieldOrdering
from ..pagination import StrawberryDjangoPagination
from ..resolvers import django_resolver


T = TypeVar("T", bound="StrawberryDjangoFieldBase")


class StrawberryDjangoFieldBase:
    def get_queryset(self, queryset: models.QuerySet, info: Info, **kwargs):
        return queryset

    @property
    def django_model(self):
        type_ = utils.unwrap_type(self.type)
        return utils.get_django_model(type_)

    @classmethod
    def from_field(cls: Type[T], field, django_type) -> T:
        raise NotImplementedError


class StrawberryDjangoField(
    StrawberryDjangoPagination,
    StrawberryDjangoFieldOrdering,
    StrawberryDjangoFieldFilters,
    StrawberryDjangoFieldBase,
    StrawberryField,
):
    """Basic field

    StrawberryDjangoField inherits all features from StrawberryField and
    implements Django specific functionalities like ordering, filtering and
    pagination.

    This field takes care of that Django ORM is always accessed from sync
    context. Resolver function is wrapped in sync_to_async decorator in async
    context. See more information about that from Django documentation.
    https://docs.djangoproject.com/en/3.2/topics/async/

    StrawberryDjangoField has following properties
    * django_name - django name which is used to access the field of model instance
    * is_auto - True if original field type was auto
    * is_relation - True if field is resolving django model relationship
    * origin_django_type - pointer to the origin of this field
    * input_type - input_type of this field used by mutations

    kwargs argument is passed to ordering, filtering, pagination and
    StrawberryField super classes.
    """

    def __init__(
        self,
        django_name: Optional[str] = None,
        graphql_name: Optional[str] = None,
        python_name: Optional[str] = None,
        **kwargs
    ):
        self.django_name = django_name
        self.is_auto = utils.is_auto(kwargs.get("type_annotation", None))
        self.is_relation = False
        self.origin_django_type = None
        self.input_type = None  # used by mutations
        super().__init__(graphql_name=graphql_name, python_name=python_name, **kwargs)

    @property
    def is_optional(self):
        return isinstance(self.type, StrawberryOptional)

    @property
    def is_list(self):
        return isinstance(self.type, StrawberryList) or (
            self.is_optional and isinstance(self.type.of_type, StrawberryList)
        )

    @classmethod
    def from_field(cls, field: "StrawberryDjangoField", django_type):
        if utils.is_strawberry_django_field(field) and not field.origin_django_type:
            return field

        new_field = StrawberryDjangoField(
            python_name=field.name,
            graphql_name=getattr(field, "graphql_name", None),
            type_annotation=field.type_annotation
            if hasattr(field, "type_annotation")
            else StrawberryAnnotation(field.type),
            description=getattr(field, "description", None),
            base_resolver=getattr(field, "base_resolver", None),
            permission_classes=getattr(field, "permission_classes", []),
            default=getattr(field, "default", UNSET),
            default_factory=field.default_factory,
            deprecation_reason=getattr(field, "deprecation_reason", None),
            directives=getattr(field, "directives", []),
            django_name=getattr(field, "django_name", field.name),
        )
        new_field.is_auto = getattr(field, "is_auto", False)
        new_field.origin_django_type = getattr(field, "origin_django_type", None)
        return new_field

    def get_result(self, source, info, args, kwargs):
        if self.base_resolver:
            return super().get_result(source, info, args, kwargs)
        return self.get_django_result(source, info, args, kwargs)

    @django_resolver
    def get_django_result(
        self,
        source,
        info,
        args,
        kwargs,
    ):
        return self.resolver(info=info, source=source, *args, **kwargs)

    def resolver(self, info, source, **kwargs):
        if source is None:
            # TODO: would there be better and safer way to detect root?
            # root query object
            result = self.django_model.objects.all()

        else:
            # relation model field
            result = getattr(source, self.django_name or self.python_name)

            if isinstance(result, models.manager.Manager):
                result = result.all()
            elif callable(result):
                result = result()

        if isinstance(result, models.QuerySet):
            result = self.get_queryset(queryset=result, info=info, **kwargs)

            if not self.is_list:
                result = result.get()

        return result

    def get_queryset(self, queryset, info, order=UNSET, **kwargs):
        type_ = self.type or self.child.type
        type_ = utils.unwrap_type(type_)
        get_queryset = getattr(type_, "get_queryset", None)
        if get_queryset:
            queryset = get_queryset(queryset, info, **kwargs)
        return super().get_queryset(queryset, info, order=order, **kwargs)

    @property
    def is_basic_field(self) -> bool:
        """
        All StrawberryDjango fields define a custom resolver that needs to be
        run, so always return False here.
        """
        return False


def field(
    resolver=None, *, name=None, field_name=None, filters=UNSET, default=UNSET, **kwargs
) -> Any:
    field_ = StrawberryDjangoField(
        python_name=None,
        graphql_name=name,
        type_annotation=None,
        filters=filters,
        django_name=field_name,
        default=default,
        **kwargs,
    )
    if resolver:
        resolver = django_resolver(resolver)
        return field_(resolver)
    return field_



from typing import List

from django.db import transaction
from strawberry import UNSET
from strawberry.annotation import StrawberryAnnotation
from strawberry.arguments import StrawberryArgument
from strawberry.type import StrawberryList, StrawberryOptional

from .. import utils
from ..fields.field import (
    StrawberryDjangoFieldBase,
    StrawberryDjangoFieldFilters,
    StrawberryField,
)
from ..fields.types import ManyToManyInput, ManyToOneInput, OneToManyInput
from ..resolvers import django_resolver


class DjangoMutationBase:
    def __init__(self, input_type, **kwargs):
        self.input_type = input_type
        super().__init__(
            graphql_name=None, python_name=None, type_annotation=None, **kwargs
        )

    @property
    def is_optional(self):
        return isinstance(self.type, StrawberryOptional)

    @property
    def is_list(self):
        return isinstance(self.type, StrawberryList) or (
            self.is_optional and isinstance(self.type.of_type, StrawberryList)
        )

    @property
    def arguments(self):
        if self.input_type:
            assert self.django_model == utils.get_django_model(
                self.input_type
            ), "Input and output types should be from the same Django model"

        arguments = []
        if self.input_type:
            is_list = self.is_list and isinstance(self, DjangoCreateMutation)
            arguments.append(get_argument("data", self.input_type, is_list))
        return arguments + super().arguments

    @django_resolver
    def get_result(self, source, info, args, kwargs):
        return self.resolver(info=info, source=source, *args, **kwargs)

    @property
    def is_basic_field(self) -> bool:
        """
        Since all mutations define a resolver function they are never basic
        fields so always return False here.
        """
        return False


class DjangoCreateMutation(
    DjangoMutationBase, StrawberryDjangoFieldBase, StrawberryField
):
    def create(self, data):
        input_data = get_input_data(self.input_type, data)
        instance = self.django_model.objects.create(**input_data)
        update_m2m([instance], data)
        return instance

    @transaction.atomic
    def resolver(self, info, source, data):
        if self.is_list:
            return [self.create(d) for d in data]
        else:
            return self.create(data)


class DjangoUpdateMutation(
    DjangoMutationBase,
    StrawberryDjangoFieldFilters,
    StrawberryDjangoFieldBase,
    StrawberryField,
):
    @transaction.atomic
    def resolver(self, info, source, data, **kwargs):
        queryset = self.django_model.objects.all()
        queryset = self.get_queryset(queryset=queryset, info=info, data=data, **kwargs)
        input_data = get_input_data(self.input_type, data)
        queryset.update(**input_data)
        update_m2m(queryset, data)
        return queryset


class DjangoDeleteMutation(
    DjangoMutationBase,
    StrawberryDjangoFieldFilters,
    StrawberryDjangoFieldBase,
    StrawberryField,
):
    @transaction.atomic
    def resolver(self, info, source, **kwargs):
        queryset = self.django_model.objects.all()
        queryset = self.get_queryset(queryset=queryset, info=info, **kwargs)
        instances = list(queryset)
        queryset.delete()
        return instances


def get_argument(name, type_, is_list=False):
    if is_list:
        return StrawberryArgument(
            python_name=name,
            graphql_name=name,
            type_annotation=StrawberryAnnotation(List[type_]),
        )
    else:
        return StrawberryArgument(
            python_name=name,
            graphql_name=name,
            type_annotation=StrawberryAnnotation(type_),
        )


def get_input_data(input_type, data):
    input_data = {}
    for field in input_type._type_definition.fields:
        value = getattr(data, field.name)
        if isinstance(value, (ManyToOneInput, ManyToManyInput)):
            continue
        if value is UNSET:
            continue
        if isinstance(value, OneToManyInput):
            value = value.set
        input_data[field.django_name] = value
    return input_data


def update_m2m(queryset, data):
    for field_name, field_value in vars(data).items():
        if not isinstance(field_value, (ManyToOneInput, ManyToManyInput)):
            continue

        for instance in queryset:
            f = getattr(instance, field_name)
            if field_value.set is not UNSET:
                if field_value.add:
                    raise ValueError("'add' cannot be used together with 'set'")
                if field_value.remove:
                    raise ValueError("'remove' cannot be used together with 'set'")

                values = field_value.set
                if isinstance(field_value, ManyToOneInput):
                    values = [f.model.objects.get(pk=pk) for pk in values]
                if values:
                    f.set(values)
                else:
                    f.clear()
            else:
                if field_value.add:
                    values = field_value.add
                    if isinstance(field_value, ManyToOneInput):
                        values = [f.model.objects.get(pk=pk) for pk in values]
                    f.add(*values)
                if field_value.remove:
                    values = field_value.remove
                    if isinstance(field_value, ManyToOneInput):
                        values = [f.model.objects.get(pk=pk) for pk in values]
                    f.remove(*values)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['strawberry_django',
 'strawberry_django.auth',
 'strawberry_django.extensions',
 'strawberry_django.fields',
 'strawberry_django.mutations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.0', 'strawberry-graphql>=0.139.0']

setup_kwargs = {
    'name': 'strawberry-graphql-django',
    'version': '0.8.2',
    'description': 'Strawberry GraphQL Django extension',
    'long_description': '# Strawberry GraphQL Django integration\n\n[![CI](https://github.com/la4de/strawberry-graphql-django/actions/workflows/main.yml/badge.svg)](https://github.com/la4de/strawberry-graphql-django/actions/workflows/main.yml)\n[![PyPI](https://img.shields.io/pypi/v/strawberry-graphql-django)](https://pypi.org/project/strawberry-graphql-django/)\n[![Downloads](https://pepy.tech/badge/strawberry-graphql-django)](https://pepy.tech/project/strawberry-graphql-django)\n\nThis package provides powerful tools to generate GraphQL types, queries, mutations and resolvers from Django models.\n\nInstalling `strawberry-graphql-django` package from the python package repository.\n\n```shell\npip install strawberry-graphql-django\n```\n\nFull documentation is available under [docs](https://strawberry-graphql.github.io/strawberry-graphql-django/) github folder.\n\n* [x] GraphQL type generation from models\n* [x] Filtering, pagination and ordering\n* [x] Basic create, retrieve, update and delete (CRUD) types and mutations\n* [x] Basic Django auth support, current user query, login and logout mutations\n* [x] Django sync and async views\n* [x] Unit test integration\n\n## Basic Usage\n\n```python\n# models.py\nfrom django.db import models\n\nclass Fruit(models.Model):\n    """A tasty treat"""\n    name = models.CharField(max_length=20)\n    color = models.ForeignKey(\'Color\', blank=True, null=True,\n            related_name=\'fruits\', on_delete=models.CASCADE)\n\nclass Color(models.Model):\n    name = models.CharField(\n        max_length=20,\n        help_text="field description",\n    )\n```\n\n```python\n# types.py\nimport strawberry\nfrom strawberry import auto\nfrom typing import List\nfrom . import models\n\n@strawberry.django.type(models.Fruit)\nclass Fruit:\n    id: auto\n    name: auto\n    color: \'Color\'\n\n@strawberry.django.type(models.Color)\nclass Color:\n    id: auto\n    name: auto\n    fruits: List[Fruit]\n```\n\n```python\n# schema.py\nimport strawberry\nfrom typing import List\nfrom .types import Fruit\n\n@strawberry.type\nclass Query:\n    fruits: List[Fruit] = strawberry.django.field()\n\nschema = strawberry.Schema(query=Query)\n```\n\nCode above generates following schema.\n\n```schema\n"""\nA tasty treat\n"""\ntype Fruit {\n  id: ID!\n  name: String!\n  color: Color\n}\n\ntype Color {\n  id: ID!\n  """\n  field description\n  """\n  name: String!\n  fruits: [Fruit!]\n}\n\ntype Query {\n  fruits: [Fruit!]!\n}\n```\n\n```python\n# urls.py\nfrom django.urls import include, path\nfrom strawberry.django.views import AsyncGraphQLView\nfrom .schema import schema\n\nurlpatterns = [\n    path(\'graphql\', AsyncGraphQLView.as_view(schema=schema)),\n]\n```\n',
    'author': 'Lauri Hintsala',
    'author_email': 'lauri.hintsala@verkkopaja.fi',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/strawberry-graphql/strawberry-graphql-django',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import asyncio
import dataclasses
import sys
import warnings
from typing import Dict

from django.db import models
from strawberry.annotation import StrawberryAnnotation
from strawberry.field import StrawberryField
from strawberry.type import StrawberryContainer


__all__ = ["deprecated"]


def is_async() -> bool:
    # django uses the same method to detect async operation
    # https://github.com/django/django/blob/bb076476cf560b988f8d80dbbc4a3c85df54b1b9/django/utils/asyncio.py
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return False
    else:
        return True


def deprecated(msg, stacklevel=1):
    warnings.warn(msg, DeprecationWarning, stacklevel=stacklevel + 1)


def is_strawberry_type(obj):
    return hasattr(obj, "_type_definition")


def is_strawberry_field(obj):
    return isinstance(obj, StrawberryField)


def is_strawberry_django_field(obj):
    from strawberry_django.fields.field import StrawberryDjangoFieldBase

    return isinstance(obj, StrawberryDjangoFieldBase)


def is_django_type(obj):
    return hasattr(obj, "_django_type")


def is_django_model(obj):
    return isinstance(obj, models.base.ModelBase)


def is_field(obj):
    return isinstance(obj, dataclasses.Field)


def is_django_field(obj):
    from .fields.field import DjangoField

    return isinstance(obj, DjangoField)


def fields(obj):
    return obj._type_definition.fields


def is_auto(obj):
    from .fields.types import is_auto

    return is_auto(obj)


def get_django_model(type_):
    if not is_django_type(type_):
        return
    return type_._django_type.model


def is_similar_django_type(a, b):
    if not a or not b:
        return False
    if a.is_input != b.is_input:
        return False
    if a.is_filter != b.is_filter:
        return False
    return True


def get_annotations(cls):
    annotations: Dict[str, StrawberryAnnotation] = {}
    namespace = sys.modules[cls.__module__].__dict__
    for c in reversed(cls.__mro__):
        if "__annotations__" in c.__dict__:
            annotations.update(
                {
                    k: StrawberryAnnotation(v, namespace=namespace)
                    for k, v in c.__annotations__.items()
                }
            )
    return annotations


def unwrap_type(type_):
    while isinstance(type_, StrawberryContainer):
        type_ = type_.of_type

    return type_
